/*
    instr16 cpu simulator
    Copyright (C) 2013  Niels Möller

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include <assert.h>
#include <math.h>
#include <stdlib.h>
#include <string.h>

#include "cpu.h"

#include "memory.h"
#include "misc.h"

void
cpu_init (struct cpu *cpu)
{
  memset (cpu->reg, 0, sizeof(cpu->reg));
  cpu->prefix_flag = 0;
  cpu->cc_flag = 0;
  cpu->trace = 0;
}

void
cpu_trace (struct cpu *cpu)
{
  cpu->trace = 1;
}

void
cpu_write (const struct cpu *cpu, FILE *f)
{
  unsigned i;
  for (i = 0; i < 16; i++)
    fprintf (f, "r%02d: %jx %ju %jd\n", i,
	     (uintmax_t) cpu->reg[i],
	     (uintmax_t) cpu->reg[i],
	     (intmax_t) (int64_t) cpu->reg[i]);
  fprintf (f, "cc: %s\n", cpu->cc_flag ? "true" : "false");
}

static void
i_ld_st_offset (struct cpu *cpu, struct memory *memory, uint16_t instr);
static void
i_ld_st_index (struct cpu *cpu, struct memory *memory, uint16_t instr);
static void
i_umull (struct cpu *cpu, uint16_t instr);
static void
i_shiftl (struct cpu *cpu, uint16_t instr);
static void
i_imm (struct cpu *cpu, uint16_t instr);
static void
i_imm_add (struct cpu *cpu, uint16_t instr);
static void
i_imm_op (struct cpu *cpu, uint16_t instr);
static void
i_imm_shift (struct cpu *cpu, uint16_t instr);
static void
i_branch (struct cpu *cpu, uint16_t instr);
static void
i_add_op (struct cpu *cpu, uint16_t instr);
static void
i_binary_op (struct cpu *cpu, struct memory *memory, uint16_t instr);
static void
i_imm_cmpgt (struct cpu *cpu, uint16_t instr);
static void
i_xshift (struct cpu *cpu, uint16_t instr);
static void
i_adcz (struct cpu *cpu, uint16_t instr);
static void
i_unary_op (struct cpu *cpu, uint16_t instr);
static void
i_nullary_op (struct cpu *cpu, uint16_t instr);
static void
i_fmac (struct cpu *cpu, uint16_t instr);
static void
i_fldexp (struct cpu *cpu, uint16_t instr);
static void
i_binary_fop (struct cpu *cpu, uint16_t instr);
static void
i_unary_fop (struct cpu *cpu, uint16_t instr);

int
cpu_instruction_cycle (struct cpu *cpu, struct memory *memory)
{
  uint16_t instr = memory_read (memory, cpu->program_counter) >> 48;
  if (cpu->trace)
    fprintf (stderr, "cpu: %4jx %04x\n",
	     (uintmax_t) cpu->program_counter, instr);
	     
  cpu->program_counter += 2;

  int prefix_flag = 0;

  if (instr < 0x4000)
    i_ld_st_offset (cpu, memory, instr);
  else if (instr < 0x5000)
    i_ld_st_index (cpu, memory, instr);
  else if (instr < 0x6000)
    i_umull (cpu, instr);
  else if (instr < 0x7000)
    i_shiftl (cpu, instr);
  else if (instr < 0x8000)
    {
      i_imm (cpu, instr);
      prefix_flag = 1;
    }
  else if (instr < 0x9000)
    i_branch (cpu, instr);
  else if (instr < 0xa000)
    i_imm_shift (cpu, instr);
  else if (instr < 0xb000)
    i_imm_add (cpu, instr);
  else if (instr < 0xc000)
    die ("old rsb instruction, not supported.\n");
  else if (instr < 0xd000)
    i_imm_op (cpu, instr);
  else if (instr < 0xe000)
    i_add_op (cpu, instr);
  else if (instr < 0xf400)
    i_binary_op (cpu, memory, instr);
  else if (instr < 0xf600)
    i_imm_cmpgt (cpu, instr);
  else if (instr < 0xf680)
    i_adcz (cpu, instr);
  else if (instr < 0xf6c0)
    i_xshift (cpu, instr);
  else if (instr < 0xf7f0)
    i_unary_op (cpu, instr);
  else if (instr < 0xf800)
    i_nullary_op (cpu, instr);
  else if (instr < 0xfa00)
    i_fmac (cpu, instr);
  else if (instr < 0xfa80)
    i_fldexp (cpu, instr);
  else if (instr < 0xfc40)
    i_binary_fop (cpu, instr);
  else if (instr < 0xfe00)
    i_unary_fop (cpu, instr);
  else if (instr == 0xfffe) /* bkpt instruction */
    cpu_write (cpu, stderr);
  else if (instr == 0xffff) /* halt instruction */
    return 0;
  else
    die ("Reserved opcode area.\n");

  cpu->prefix_flag = prefix_flag;
  return 1;
}

static double
get_fpreg (const struct cpu *cpu, unsigned i)
{
  union {
    uint64_t i;
    double f;
  } u;
  u.i = cpu->reg[i];
  return u.f;
}

static void
set_fpreg (struct cpu *cpu, unsigned i, double f)
{
  union {
    uint64_t i;
    double f;
  } u;
  u.f = f;
  cpu->reg[i] = u.i;
}

/* Single precision representation */
static double
get_fpreg_s (const struct cpu *cpu, unsigned i)
{
  union {
    uint32_t i;
    float f;
  } u;
  u.i = cpu->reg[i];
  return u.f;
}

static void
set_fpreg_s (struct cpu *cpu, unsigned i, float f)
{
  union {
    uint32_t i;
    float f;
  } u;
  u.f = f;
  cpu->reg[i] = u.i;
}


static uint64_t
imm4 (const struct cpu *cpu, unsigned char imm)
{
  static const uint8_t imm4_table[16] =
    {
      32, 1,  2,  3,  4,  5,  6,  7,
      8, 10, 12, 14, 16, 20, 24, 28
    };
  return cpu->prefix_flag
    ? imm | (cpu->prefix_reg << 4)
    : imm4_table[imm];
}

static uint64_t
imm_cmp (unsigned char imm)
{
  static const uint8_t imm_cmp_table[16] =
    {
      32, 10, 12, 14, 16, 20, 24, 28,
    };
  uint64_t x = imm_cmp_table[imm & 7];
  if (imm >> 3)
    x = - x;
  return x;  
}

/* Imm-value used for branches */
static uint64_t
imm9 (const struct cpu *cpu, unsigned imm)
{
  return cpu->prefix_flag ? 
    imm | (cpu->prefix_reg << 9)
    : imm;
}

static void
umul64 (uint64_t *hp, uint64_t *lp, uint64_t a, uint64_t b)
{
  uint64_t p0, m0, m1, p1;
  uint64_t ah, al, bh, bl;

  ah = a >> 32;
  al = a & 0xffffffffUL;
  bh = b >> 32;
  bl = b & 0xffffffffUL;

  p0 = al*bl;
  m0 = al*bh;
  m1 = ah*bl;
  p1 = ah*bh;

  m0 += m1;
  p1 += (uint64_t) (m0 < m1) << 32;
  p1 += m0 >> 32;
  m0 <<= 32;
  p0 += m0;
  p1 += (p0 < m0);
  *hp = p1;
  if (lp)
    *lp = p0;
}

static uint64_t
addc (struct cpu *cpu, uint64_t a, uint64_t b, uint64_t cy, unsigned char mode)
{
  uint64_t sum;
  sum = a + cy;
  cy = sum < a;
  sum += b;
  cy += sum < b;

  if (mode & 2)
    {
      /* Signed */
      /* If we follow the ARM spec (A2-43), the "true result" is a + b
	 + cy, where a and b are interpreted as signed, and cy is
	 unsigned. We still never get overflow if a and b are of
	 opposite sign, so the overflow condition is the same: that a
	 and b have the same sign, and the result has the opposite
	 sign. */

      /* For adds (mode == 2), cc is set if a >= b (signed), i.e., if
	 the true signed result is >= 0. */

      int asign = a >> 63;
      int bsign = b >> 63;
      int ssign = sum >> 63;

      /* Overflow */
      cpu->cc_flag = (asign ^ bsign ^ 1) & (asign ^ ssign);
      if ( (mode & 1) == 0)
	cpu->cc_flag ^= (ssign ^ 1);
    }
  else
    if (mode & 1)
      cpu->cc_flag = cy;

  return sum;
}

static uint64_t
clz (uint64_t x)
{
  if (x == 0)
    return 64;
  else
    {
      unsigned cnt;
      for (cnt = 0;
	   (x & ((uint64_t) 1 << 63)) == 0;
	   cnt++, x <<= 1)
	;
      return cnt;
    }
}

static uint64_t
ctz (uint64_t x)
{
  if (x == 0)
    return 64;
  else
    {
      unsigned cnt;
      for (cnt = 0; (x & 1) == 0; cnt++, x >>= 1)
	;
      return cnt;
    }
}

static unsigned
cls (uint64_t x)
{
  if (x & ((uint64_t) 1 << 63))
    x = ~x;
  return clz (x);
}

static unsigned
popc (uint64_t x)
{
  unsigned cnt, i;
  for (cnt = i = 0; i < 64; i++, x >>= 1)
    cnt += x & 1;
  return cnt;
}

static uint64_t
bswap (uint64_t x)
{
  uint64_t y;
  unsigned i;
  for (i = y = 0; i < 8; i++, x >>= 8)
    y = (y << 8) | (x & 0xff);
  return y;
}

static uint64_t
udiv_21 (uint64_t u1, uint64_t u0, uint64_t d)
{
  uint64_t dh, dl, qh, ql, p;
  assert (d & ((uint64_t) 1 << 63));
  assert (u1 < d);
  dh = d >> 32;
  dl = d - (dh << 32);
  
  qh = u1 / (dh + 1);

  u1 -=  qh * dh;
  p = qh * dl;
  u1 -= p >> 32;
  p <<= 32;
  u1 -= u0 < p;
  u0 -= p;

  while (u1 >= ((uint64_t) 1 << 32))
    {
      qh++;
      u1 -= dh + (u0 < (dl << 32));
      u0 -= (dl << 32);
    }
  ql = ((u1 << 32) + (u0 >> 32)) / (dh+1);
  
  p = ql * dl;
  u1 -= (u0 < p);
  u0 -= p;
  p = ql * dh;
  u1 -= p >> 32;
  p <<= 32;
  u1 -= (u0 < p);
  u0 -= p;
  
  while (u1 > 0)
    {
      ql++;
      u1 -= u0 < d;
      u0 -= d;
    }

  ql += (u0 >= d);

  return (qh << 32) + ql;
}

static void
i_ld_st_offset (struct cpu *cpu, struct memory *memory, uint16_t instr)
{
  /* Load and store with offset. */
  unsigned op = instr >> 12;
  unsigned s = (instr >> 4) & 15;
  unsigned d = instr & 15;
  uint64_t offset = imm4 (cpu, (instr >> 8) & 15);
  uint64_t ea;

  if (op & 1)
    ea = cpu->reg[s] - offset;
  else
    ea = cpu->reg[s] + offset;

  if (op & 2)
    memory_write (memory, ea, cpu->reg[d]);
  else
    cpu->reg[d] = memory_read (memory, ea);
}

static void
i_ld_st_index (struct cpu *cpu, struct memory *memory, uint16_t instr)
{
  unsigned s0 = (instr >> 8) & 15;
  unsigned s1 = (instr >> 4) & 15;
  uint64_t ea = cpu->reg[s0] + cpu->reg[s1];
  unsigned d = instr & 15;
  if (s0 == s1)
    die ("Invalid indexed memory access instruction.\n");
  if (s1 < s0)
    cpu->reg[d] = memory_read (memory, ea);
  else
    memory_write (memory, ea, cpu->reg[d]);
}

static void
i_umull (struct cpu *cpu, uint16_t instr)
{
  unsigned h = (instr >> 8) & 15;
  unsigned a = (instr >> 4) & 15;
  unsigned d = instr & 15;
  umul64 (&cpu->reg[h], &cpu->reg[d],
	  cpu->reg[a],cpu->reg[d]);
}

static void
i_shiftl (struct cpu *cpu, uint16_t instr)
{
  unsigned b = (instr >> 8) & 15;
  unsigned c = (instr >> 4) & 15;
  unsigned d = instr & 15;
  int64_t shift = cpu->reg[c];
  /* FIXME: Handle the range 64 < |shift| < 128? */
  if (shift > 0)
    {
      if (shift >= 64)
	cpu->reg[d] = 0;
      else if (b == 15)
	cpu->reg[d]
	  = (cpu->reg[d] << shift) | ((uint64_t) cpu->cc_flag << (shift - 1));
      else
	cpu->reg[d]
	  = (cpu->reg[d] << shift)
	  | (cpu->reg[b] >> (64 - shift));
    }
  else
    {
      if (shift <= -64)
	cpu->reg[d] = 0;
      else if (b == 15)
	cpu->reg[d]
	  = (cpu->reg[d] >> -shift) | ((uint64_t) cpu->cc_flag << (64 + shift));
      else
	cpu->reg[d]
	  = (cpu->reg[d] >> -shift)
	  | (cpu->reg[b] << (64 + shift));
    }
}

static void
i_imm (struct cpu *cpu, uint16_t instr)
{
  uint64_t prefix = instr & 0xfff;
  if (cpu->prefix_flag)
    prefix |= cpu->prefix_reg << 12;

  cpu->prefix_reg = prefix;
}

static void
i_imm_add (struct cpu *cpu, uint16_t instr)
{
  uint64_t imm = imm4 (cpu, (instr >> 4) & 15);
  unsigned sign = (instr >> 8) & 1;
  unsigned d = instr & 15;
  /* FIXME: For immediate sub cc, we need one's complement. */
  if (sign)
    imm = -imm;
  
  cpu->reg[d] = addc (cpu, cpu->reg[d], imm,
		      (instr>>11) & cpu->cc_flag,
		      (instr >> 9) & 3);
}

static void
i_imm_op (struct cpu *cpu, uint16_t instr)
{
  uint64_t imm = imm4 (cpu, (instr >> 4) & 15);
  unsigned sign = (instr >> 8) & 1;
  unsigned d = instr & 15;

  if (sign)
    imm = ~imm;

  switch ( (instr >> 9) & 7)
    {
    case 0: /* mov imm */
      cpu->reg[d] = imm + sign;
      break;
    case 1: /* and imm */
      cpu->reg[d] &= imm;
      break;
    case 2: /* or imm */
      cpu->reg[d] |= imm;
      break;
    case 3: /* xor imm */
      cpu->reg[d] ^= imm;
      break;
    case 4: /* tst imm */
      cpu->cc_flag = (cpu->reg[d] & imm) != 0;
      break;
    case 5: /* cmpeq imm */
      cpu->cc_flag = (cpu->reg[d] == imm + sign);
      break;
    case 6: /* cmpugeq imm */
      imm += sign;
      if (cpu->prefix_flag)
	goto cmpugeq;
      /* Steal some codes */
      switch (imm)
	{
	case - (uint64_t) 1:
	  cpu->cc_flag = ( (int64_t) cpu->reg[d] >= 0);
	  break;
	case 2:
	  cpu->cc_flag = (cpu->reg[d] == 0);
	  break;
	case 4:
	  cpu->cc_flag = ( (int64_t) cpu->reg[d] > 8);
	  break;
	case 8:
	  cpu->cc_flag = (cpu->reg[d] > 8U);
	  break;
	default:
	cmpugeq:
	  cpu->cc_flag = (cpu->reg[d] >= imm);
	}
      break;
    case 7: /* cmpsgeq imm */
      imm += sign;
      cpu->cc_flag = ( (int64_t) cpu->reg[d] >= (int64_t) imm);
      break;
    }
}

static void
i_imm_shift (struct cpu *cpu, uint16_t instr)
{
  unsigned shift = (instr >> 4) & 63;
  unsigned d = instr & 15;
  unsigned op = (instr >> 10) & 3;

  if (shift > 0)
    switch (op)
      {
      case 0: /* lshift */
	cpu->reg[d] <<= shift;
	break;
      case 1: /* rshift */
	cpu->reg[d] >>= shift;
	break;
      case 2: /* ashift */
	{
	  /* Arithmetic shift */
	  uint64_t sign = - (cpu->reg[d] >> 63);
	  cpu->reg[d]
	    = (cpu->reg[d] >> shift) | (sign << (64 - shift));
	  break;
	}
      case 3: /* rot left */
	cpu->reg[d] = (cpu->reg[d] << shift)
	  | (cpu->reg[d] >> (64 - shift));
      }
  else
    /* Zero shift count is special */
    switch (op)
      {
      case 0:
	cpu->reg[d] = clz (cpu->reg[d]);
	break;
      case 1:
	cpu->reg[d] = ctz (cpu->reg[d]);
	break;
      case 2:
	cpu->reg[d] = cls (cpu->reg[d]);
	break;
      case 3:
	cpu->reg[d] = popc (cpu->reg[d]);
      }
}

static void
i_branch (struct cpu *cpu, uint16_t instr)
{
  uint64_t imm = (imm9 (cpu, instr & 0x1ff) + 1)<< 1;
  if (instr & 0x200)
    imm = -imm;

  switch ( (instr >> 10) & 3)
    {
    case 1: /* jsr */
      cpu->reg[14] = cpu->program_counter;
      /* Fall through */
    case 0: /* jmp */
      cpu->program_counter += imm;
      break;
    case 2: /* bt */
      if (cpu->cc_flag)
	cpu->program_counter += imm;
      break;
    case 3: /* bf */
      if (!cpu->cc_flag)
	cpu->program_counter += imm;
      break;
    }
}

static void
i_add_op (struct cpu *cpu, uint16_t instr)
{
  unsigned dreg = instr & 15;
  uint64_t s = cpu->reg[(instr >> 4) & 15];
  uint64_t sub_flag = (instr >> 11) & 1;
  uint64_t cin_flag = ((instr >> 10) & 1);
  uint64_t cy;

  s ^= -sub_flag;
  cy = cin_flag ? cpu->cc_flag : sub_flag;

  cpu->reg[dreg] = addc (cpu, cpu->reg[dreg], s, cy, (instr >> 8) & 3);
}

static void
i_binary_op (struct cpu *cpu, struct memory *memory, uint16_t instr)
{
  unsigned dreg = instr & 15;
  uint64_t s = cpu->reg[(instr >> 4) & 15];
  switch ( (instr >> 8) & 31)
    {
    case 0: /* mov */
      cpu->reg[dreg] = s;
      break;
    case 1: /* movt */
      if (cpu->cc_flag)
	cpu->reg[dreg] = s;
      break;
    case 2: /* movf */
      if (!cpu->cc_flag)
	cpu->reg[dreg] = s;
      break;
    case 3: /* and */
      cpu->reg[dreg] &= s;
      break;
    case 4: /* or */
      cpu->reg[dreg] |= s;
      break;
    case 5: /* xor */
      cpu->reg[dreg] ^= s;
      break;
    case 6: /* mullo */
      cpu->reg[dreg] *= s;
      break;
    case 7: /* umulhi */
      umul64 (&cpu->reg[dreg], NULL, cpu->reg[dreg], s);
      break;
    case 8: /* lshift */
      {
	int64_t cnt = s;
	if (cnt >= 64 || cnt <= -64)
	  cpu->reg[dreg] = 0;
	else if (cnt > 0)
	  cpu->reg[dreg] <<= cnt;
	else
	  cpu->reg[dreg] >>= -cnt;
	break;
      }
    case 9: /* ashift */
      {
	int64_t cnt = s;
	if (cnt >= 0)
	  cpu->reg[dreg] = ((cnt < 64) ? (cpu->reg[dreg] << cnt) : 0);
	else
	  {
	    /* Arithmetic shift */
	    uint64_t sign = - (cpu->reg[dreg] >> 63);
	    cpu->reg[dreg]
	      = (cnt > -64)
	      ? (cpu->reg[dreg] >> - cnt) | (sign << (64 + cnt))
	      : sign;
	  }
	break;
      }
    case 10: /* injt8 */
      cpu->reg[dreg] = (cpu->reg[dreg] & (((uint64_t) 1 << 56)-1))
	| s << 56;
      break;
    case 11: /* injt16 */
      cpu->reg[dreg] = (cpu->reg[dreg] & (((uint64_t) 1 << 48)-1))
	| s << 48;
      break;
    case 12: /* injt32 */
      cpu->reg[dreg] = (cpu->reg[dreg] & (((uint64_t) 1 << 32)-1))
	| s << 32;
      break;
    case 13: /* tst */
      cpu->cc_flag = (cpu->reg[dreg] & s) != 0;
      break;
    case 14: /* cmpeq */
      cpu->cc_flag = (cpu->reg[dreg] == s);
      break;
    case 15: /* cmpugeq */
      cpu->cc_flag = (cpu->reg[dreg] >=  s);
      break;
    case 16: /* cmpuleq */
      cpu->cc_flag = ((int64_t) cpu->reg[dreg] >= (int64_t) s);
      break;
    case 17: /* ld */
      cpu->reg[dreg] = memory_read (memory, s);
      break;
    case 18: /* st */
      memory_write (memory, s, cpu->reg[dreg]);
      break;
    default:
      die ("Unassigned instr.\n");
    }
}

static void
i_imm_cmpgt (struct cpu *cpu, uint16_t instr)
{
  unsigned dreg = instr & 15;
  uint64_t imm = imm_cmp ( (instr >> 4) & 15);
  if ((instr >> 8) & 1)
    /* Signed cmp */
    cpu->cc_flag = (int64_t) cpu->reg[dreg] > (int64_t) imm;
  else
    /* Unsigned cmp */
    cpu->cc_flag = cpu->reg[dreg] > imm;
}

static void
i_xshift (struct cpu *cpu, uint16_t instr)
{
  unsigned dreg = instr & 15;
  switch ((instr >> 4) & 3)
    {
    case 0:
      cpu->reg[dreg] = (cpu->reg[dreg] >> 1)
	| ( (uint64_t) cpu->cc_flag << 63);
      break;
    case 1:
      {
	uint64_t cin = cpu->cc_flag;
	cpu->cc_flag = cpu->reg[dreg] & 1;
	cpu->reg[dreg] = (cpu->reg[dreg] >> 1)
	  | (cin << 63);
	break;
      }
    case 2:
      cpu->cc_flag = cpu->reg[dreg] & 1;
      cpu->reg[dreg] >>= 1;
      break;
    case 3:
      /* Arithmetic shift */
      cpu->cc_flag = cpu->reg[dreg] & 1;
      cpu->reg[dreg] = (cpu->reg[dreg] >> 1)
	| (cpu->reg[dreg] & ((uint64_t) 1 << 63));
      break;
    }
}

static void
i_adcz (struct cpu *cpu, uint16_t instr)
{
  unsigned dreg = instr & 15;
  uint64_t sign = (instr >> 6) & 1;
  cpu->reg[dreg] = addc (cpu, cpu->reg[dreg], -sign,
			 cpu->cc_flag, (instr >> 4) & 3);
}

static void
i_unary_op (struct cpu *cpu, uint16_t instr)
{
  unsigned dreg = instr & 15;
  switch ((instr >> 4) & 7)
    {
    case 4: /* neg */
      cpu->reg[dreg] = - cpu->reg[dreg];
      break;
    case 5: /* not */
      cpu->reg[dreg] = ~cpu->reg[dreg];
      break;
    case 6: /* bswap */
      cpu->reg[dreg] = bswap (cpu->reg[dreg]);
      break;
    case 7: /* recpr */
      {
	uint64_t d = cpu->reg[dreg];
	if (d == 0)
	  die ("Divide by zero.\n");
	d <<= clz (d);
	/* Compute floor [(2^128 - 1) / d] - 2^64
	   = floor [<2^64 - 1 - d, 2^64 - 1> / d] */
	cpu->reg[dreg] = udiv_21 (~d, ~(uint64_t) 0, d);
	break;
      }
    case 0: /* call, indirect */
      cpu->reg[14] = cpu->program_counter;
      cpu->program_counter = cpu->reg[dreg];
      break;
    default:
      die ("Unassigned opcode area.\n");
    }
}

static void
i_nullary_op (struct cpu *cpu, uint16_t instr)
{
  if (instr == 0xf7f80) /* not cc */
    cpu->cc_flag = ! cpu->cc_flag;
  else
    die ("Invlid instruction.\n");
}

static void
i_fmac (struct cpu *cpu, uint16_t instr)
{
  unsigned dreg = instr & 7;
  double d = get_fpreg (cpu, dreg);

  d += get_fpreg (cpu, (instr >> 6)  & 7)
    * get_fpreg (cpu, (instr >> 3)  & 7);
  set_fpreg (cpu, dreg, d);
}

static void
i_fldexp (struct cpu *cpu, uint16_t instr)
{
  unsigned dreg = instr & 7;
  double d = get_fpreg (cpu, dreg);
  set_fpreg (cpu, dreg, ldexp (d, cpu->reg[(instr >> 3) & 15]));
}

static void
i_binary_fop (struct cpu *cpu, uint16_t instr)
{
  unsigned dreg = instr & 7;
  unsigned sreg = (instr >> 3) & 7;
  double d = get_fpreg (cpu, dreg);
  double s = get_fpreg (cpu, sreg);

  switch ((instr >> 6) & 31)
    {
    case 10: /* fadd */
      set_fpreg (cpu, dreg, d + s);
      break;
    case 11: /* fsub */
      set_fpreg (cpu, dreg, d - s);
      break;
    case 12: /* fmul */
      set_fpreg (cpu, dreg, d * s);
      break;
    case 13: /* fdiv */
      set_fpreg (cpu, dreg, d / s);
      break;
    case 14: /* fcmpeq */
      cpu->cc_flag = (d == s);
      break;
    case 15: /* fcmpgeq */
      cpu->cc_flag = (d >= s);
      break;
    case 16: /* fcmpgt */
      cpu->cc_flag = (d > s);
      break;
    default: abort();
    }
}

static void
i_unary_fop (struct cpu *cpu, uint16_t instr)
{
  unsigned dreg = instr & 7;
  double d = get_fpreg (cpu, dreg);
  switch ((instr >> 3) & 31)
    {
    case 8: /* fneg */
      set_fpreg (cpu, dreg, -d);
      break;
    case 9: /* fabs */
      set_fpreg (cpu, dreg, fabs (d));
      break;
    case 10: /* fs2d */
      set_fpreg (cpu, dreg, get_fpreg_s (cpu, dreg));
      break;
    case 11: /* fd2s */
      set_fpreg_s (cpu, dreg, d);
      break;
    case 12: /* fui2d */
      set_fpreg (cpu, dreg, (double) cpu->reg[dreg]);
      break;
    case 13: /* fd2ui */
      cpu->reg[dreg] = d;
      break;
    case 14: /* fsi2d */
      set_fpreg (cpu, dreg, (double) (int64_t) cpu->reg[dreg]);
      break;
    case 15: /* fd2si */
      cpu->reg[dreg] = (int64_t) d;
      break;
    case 16: /* fui2s */
      set_fpreg_s (cpu, dreg, (float) cpu->reg[dreg]);
      break;
    case 17: /* fsi2s */
      set_fpreg_s (cpu, dreg, (float) (int64_t) cpu->reg[dreg]);
      break;
    case 18: /* feqz */
      cpu->cc_flag = (d == 0.0);
      break;
    case 19: /* fgeqz */
      cpu->cc_flag = (d >= 0.0);
      break;
    case 20: /* fgtz */
      cpu->cc_flag = (d < 0.0);
      break;
    case 21: /* fleqz */
      cpu->cc_flag = (d <= 0.0);
      break;
    case 22: /* fltz */
      cpu->cc_flag = (d < 0.0);
      break;
    default:
      die ("Unassigned opcode area.\n");
    }
}
